import itertools
from functools import cache
from typing import Any, Iterator, Literal

from ..base import BaseSolver

NUM_PAD_P = {
    v: (i, j)
    for i, r in enumerate(("789", "456", "123", " 0A"))
    for j, v in enumerate(r)
    if v.strip()
}
MOV_PAD_P = {
    v: (i, j)
    for i, r in enumerate((" ^A", "<v>"))
    for j, v in enumerate(r)
    if v.strip()
}


def path(start: tuple[int, int], end: tuple[int, int], pad: Literal["num", "mov"]):
    # a move in the grid is composed of at most two straight line: up/down and
    # left/right, since doing some kind of diagonal moves would create long path for
    # the robot above (since this involves going back-and-forth to the letter 'A')
    #
    row_s, col_s = start
    row_e, col_e = end

    le, de, ue, re = (
        "<" * max(0, col_s - col_e),
        "v" * max(0, row_e - row_s),
        "^" * max(0, row_s - row_e),
        ">" * max(0, col_e - col_s),
    )

    # when the robot starts or ends on the row/column with the empty cell, there is
    # only one way to move
    #
    if pad == "num" and (row_s, col_e) == (3, 0):
        return ue + le
    elif pad == "num" and (col_s, row_e) == (0, 3):
        return re + de
    elif pad == "mov" and col_s == 0:
        return re + ue
    elif pad == "mov" and col_e == 0:
        return de + le

    # otherwise, we need to decide if we want to go up/down first, or left/right, and
    # apparently this is the best way to do it...
    return le + de + ue + re


@cache
def v_clicks(clicks: str, depth: int) -> int:
    if depth == 0:
        return len(clicks)

    n_clicks = 0
    at = "A"
    for _, group in itertools.groupby(clicks):
        group = list(group)
        n_clicks += v_clicks(
            path(MOV_PAD_P[at], MOV_PAD_P[group[0]], "mov") + "A" * len(group),
            depth - 1,
        )
        at = group[0]

    return n_clicks


def path_length(code: str, depth: int):
    return sum(
        v_clicks(path(NUM_PAD_P[start], NUM_PAD_P[end], "num") + "A", depth)
        for start, end in zip("A" + code[:-1], code, strict=True)
    )


class Solver(BaseSolver):
    def solve(self, input: str) -> Iterator[Any]:
        yield sum(
            path_length(code, 2) * int(code[:-1], 10) for code in input.splitlines()
        )
        yield sum(
            path_length(code, 25) * int(code[:-1], 10) for code in input.splitlines()
        )
